import os
import glob
import zipfile
import shutil

import toml
from tqdm import tqdm
import trimesh
from PIL import Image
import numpy as np

from h3ds.log import logger
from h3ds.utils import download_file_from_google_drive, md5, load_K_Rt


class H3DS:

    def __init__(self, path: str):

        self.path = path

        if not os.path.exists(self.path):
            logger.warning(
                f'{self.path} does not exist. Change the path or call H3DS.download.'
            )

        self._load_config()

    def download(self, token):
        # Download zip file
        tmp_dir = os.path.join(self.path, 'tmp')
        if not os.path.exists(tmp_dir):
            os.makedirs(tmp_dir)

        tmp_zip = os.path.join(tmp_dir, 'h3ds.zip')
        logger.print(f'Downloading H3DS dataset to {tmp_zip}')
        download_file_from_google_drive(id=self._config['file_id'],
                                        destination=tmp_zip)

        # Check md5
        md5_zip = md5(tmp_zip)
        if md5_zip == self._config['file_md5']:
            logger.print('MD5 check - Success')
        else:
            logger.error('MD5 check - Failed')

        # Unzip into self.path
        logger.print(f'Unzipping file to {self.path}')
        with zipfile.ZipFile(tmp_zip, 'r') as zip_ref:
            for member in tqdm(zip_ref.infolist(), desc='Extracting...'):
                zip_ref.extract(member, self.path, pwd=token.encode('utf-8'))

        # Remove temporal zip
        logger.print(f'Removing temporary files')
        shutil.rmtree(tmp_dir)

    def scenes(self):
        return list(self._config['scenes'].keys())

    def default_views_configs(self, scene_id: str):
        return list(
            self._config['scenes'][scene_id]['default_views_configs'].keys())

    def load_scene(self, scene_id: str, views_config_id: str = None):
        scene_path = os.path.join(self.path, scene_id)

        mesh = self.load_mesh(scene_id)
        images = self.load_images(scene_id, views_config_id)
        masks = self.load_masks(scene_id, views_config_id)
        cameras = self.load_cameras(scene_id, views_config_id)

        return mesh, images, masks, cameras

    def load_mesh(self, scene_id: str):
        mesh_path = os.path.join(self.path, scene_id, 'full_head.obj')

        return trimesh.load(mesh_path, process=False)

    def load_images(self, scene_id: str, views_config_id: str = None):
        images = self._load_images(scene_id, 'image')

        return self._filter_views(images, scene_id, views_config_id)

    def load_masks(self, scene_id: str, views_config_id: str = None):
        masks = self._load_images(scene_id, 'mask')

        return self._filter_views(masks, scene_id, views_config_id)

    def load_cameras(self, scene_id: str, views_config_id: str = None):
        cameras_path = os.path.join(self.path, scene_id, 'cameras.npz')
        camera_dict = np.load(cameras_path)

        cameras = []
        for idx in range(self._config['scenes'][scene_id]['views']):
            s = camera_dict['scale_mat_%d' % idx].astype(np.float32)
            P = camera_dict['world_mat_%d' % idx].astype(np.float32)
            K, P = load_K_Rt(P[:3, :4])
            cameras.append((s, K, P))

        return self._filter_views(cameras, scene_id, views_config_id)

    def _load_config(self):
        h3ds_dir = os.path.dirname(os.path.realpath(__file__))
        config_path = os.path.join(h3ds_dir, 'config.toml')

        self._config = toml.load(config_path)

    def _get_views_config(self, scene_id: str, config_id: str):

        if scene_id not in self._config['scenes']:
            logger.exception(f'Invalid scene_id {scene_id}')

        if config_id not in self._config['scenes'][scene_id][
                'default_views_configs']:
            logger.exception(
                f'Invalid config_id {config_id} for scene_id {scene_id}')

        return self._config['scenes'][scene_id]['default_views_configs'][
            config_id]

    def _load_images(self, scene_id: str, rel_path: str):
        images_dir = os.path.join(self.path, scene_id, rel_path)
        images_paths = sorted(list(glob.glob(os.path.join(images_dir,
                                                          '*.jpg'))))
        return [Image.open(img).copy() for img in images_paths]

    def _filter_views(self,
                      elements,
                      scene_id: str,
                      views_config_id: str = None):

        if views_config_id is None:
            return elements

        views_config = self._get_views_config(scene_id, views_config_id)
        return [elements[idx] for idx in views_config]


import cv2
import hashlib
from requests import Session, Response
import numpy as np


def download_file_from_google_drive(id: str, destination: str):

    URL = "https://docs.google.com/uc?export=download"

    session = Session()

    response = session.get(URL, params={'id': id}, stream=True)
    token = get_confirm_token(response)

    if token:
        params = {'id': id, 'confirm': token}
        response = session.get(URL, params=params, stream=True)

    save_response_content(response, destination)


def get_confirm_token(response: Response):
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None


def save_response_content(response: Response, destination: str):
    CHUNK_SIZE = 32768

    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)


def md5(filepath: str):
    hash_md5 = hashlib.md5()
    with open(filepath, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def load_K_Rt(P: np.ndarray):

    dec = cv2.decomposeProjectionMatrix(P)
    K = dec[0]
    R = dec[1]
    t = dec[2]

    K = K / K[2, 2]
    intrinsics = np.eye(4)
    intrinsics[:3, :3] = K

    pose = np.eye(4, dtype=np.float32)
    pose[:3, :3] = R.transpose()
    pose[:3, 3] = (t[:3] / t[3])[:, 0]

    return intrinsics, pose



import os
import re
from tempfile import mkstemp
from setuptools import setup, find_packages
from pkg_resources import parse_requirements

PACKAGE_NAME = 'h3ds'
PACKAGE_PATH = PACKAGE_NAME
VERSION_FILE_PATH = os.path.join(PACKAGE_PATH, '__init__.py')
REQUIREMENTS_FILE_PATH = 'requirements.txt'

# Read __version__ from VERSION_FILE_PATH
exec(open(VERSION_FILE_PATH).read())

with open("README.md", "r", encoding="utf-8") as fh:
    LONG_DESCRIPTION = fh.read()

# Build setup
setup(
    name='h3ds',
    version=__version__,
    author='Eduard Ramon Maldonado',
    author_email='eduard.ramon.maldonado@gmail.com',
    description='Python interface for H3DS dataset',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://test.pypi.org/project/h3ds/',
    project_urls={
        'Bug Tracker': 'https://github.com/CrisalixSA/h3ds/issues',
        'Project Website': 'https://crisalixsa.github.io/h3d-net/'
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={PACKAGE_NAME: PACKAGE_PATH},
    packages=find_packages(
        include=[PACKAGE_NAME],
        exclude=[],
    ),
    install_requires=[
        'toml',
        'requests',
        'trimesh',
        'Pillow',
        'tqdm',
        'opencv-python'

    ],
)



from os import getenv
from logging import NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL
from logging import (getLogger, getLoggerClass, setLoggerClass, addLevelName,
                     StreamHandler, Formatter, _nameToLevel)

PRINT = 100
COLORS = {
    PRINT: '\u001b[32m{}\033[0m',
    CRITICAL: '\u001b[35m{}\033[0m',
    ERROR: '\u001b[31m{}\033[0m',
    WARNING: '\u001b[33m{}\033[0m',
    INFO: '\u001b[36m{}\033[0m',
    DEBUG: '\u001b[37m{}\033[0m',
}


class Logger(getLoggerClass()):
    """Custom logger class"""

    def __init__(self, name: str, *args, **kwargs) -> None:
        """Logger class constructor.

        Args:
            name (str): Name of the module.
        """
        super().__init__(name, *args, **kwargs)
        log_limit = int(getenv('LOGLIMIT', 10000)) + len(COLORS[DEBUG]) - 1
        fmt = f'%(levelname)8.8s [%(asctime)23.23s] %(message).{log_limit}s'
        self.propagate = False
        handler = StreamHandler()
        self.loadLevel()
        handler.setLevel(NOTSET)
        formatter = Formatter(fmt)
        handler.setFormatter(formatter)
        self.addHandler(handler)
        addLevelName(PRINT, 'PRINT')

    def loadLevel(self) -> None:
        """Load log level from environment variable"""
        self.setLevel(_nameToLevel.get(getenv('LOGLEVEL', 'INFO'), INFO))

    def print(self, msg: str, *args, **kwargs) -> None:
        """Log a print-level message"""
        if self.isEnabledFor(PRINT):
            self._log(PRINT, msg, args, **kwargs)

    def critical(self, msg: str, *args, **kwargs) -> None:
        """Log a critical-level message and raise and exception"""
        if self.isEnabledFor(CRITICAL):
            self._log(CRITICAL, msg, args, **kwargs)
            raise Exception(msg)

    def _log(self, level: int, msg: str, *args, **kwargs) -> None:
        """Log method wrapper for colouring"""
        super()._log(level, COLORS[level].format(msg), *args, **kwargs)


class ErrorLogger:

    def __init__(self):
        self.errors = []

    def log(self, execution):
        if execution != 0:
            self.errors.append(execution)


def set_logger():
    """Set custom logger class as the default class"""
    setLoggerClass(Logger)


def get_logger():
    """Get current package logger class"""
    set_logger()
    package = __name__.split('.')[0]
    return getLogger(package)


logger = get_logger()
